import tensorflow as tf
import pickle


def save_model_json(model, model_name):
    # serialize model to json
    json_model = model.to_json()
    # save the model architecture to JSON file
    with open(f'{model_name}.json', 'w') as json_file:
        json_file.write(json_model)
    # saving the weights of the model
    model.save_weights(f'{model_name}_weights.h5')


def load_model_json(model_name, custom_objects=None):
    """[summary]


    Args:
        model_name ([type]): [description]
        custom_objects ([dict], optional): Dictionary to pass to tf.keras.models.model_from_json.
        Use it to specify things like LeakyReLU.
        See: https://stackoverflow.com/questions/55364954/keras-load-model-cant-recognize-tensorflows-activation-functions
        See: https://github.com/tensorflow/tfjs/issues/1093
        Defaults to None.

    Returns:
        [type]: [description]
    """
    # Reading the model from JSON file
    with open(f'{model_name}.json', 'r') as json_file:
        json_savedModel = json_file.read()
    # load the model architecture
    model = tf.keras.models.model_from_json(
        json_savedModel, custom_objects=custom_objects)
    # model.summary()
    model.load_weights(f'{model_name}_weights.h5')
    return model


def save_model_h5(model, model_name):
    """save keras model as h5 file

    Args:
        model (keras model): [description]
        model_name (str): output file name will be f"{model_name}.h5"
    """
    model.save(f'{model_name}.h5')


def load_model_h5(filename, custom_objects=None):
    """load h5 model

    Args:
        filename (str): h5 file containing saved keras model
        custom_objects ([dict], optional): Dictionary to pass to tf.keras.models.load_model.
    """
    return tf.keras.models.load_model(filename, custom_objects=custom_objects)


def save_model_checkpoint(model, model_name, num, zpad='04'):
    """
    name format: '{model_name}+checkpoint_0001.h5'
    """
    # saving the smodel's architecture, weights, and training configuration in a single file/folder.
    model.save(f'{model_name}_checkpoint_' + format(num, zpad) + '.h5')


def load_model_checkpoint(model_name, num, zpad='04', custom_objects=None):
    """
    name format: '{model_name}+checkpoint_0001.h5'
    custom_objects ([dict], optional): Dictionary to pass to tf.keras.models.load_model.
    """
    # loading the model from the HDF5 file
    model = tf.keras.models.load_model(
        f'{model_name}_checkpoint_' + format(num, zpad) + '.h5', custom_objects=custom_objects)
    return model


def save_history(history, filename):
    """saves the History.history dictionary of a model.fit() to a pickle file

    Args:
        history (dict): History.history dictionary, output from model.fit().
        filename (str): full path and name output pickle file
    """

    with open(filename, 'wb') as pfile:
        pickle.dump(history, pfile)


def load_history(filename):
    """load the history.history dictionary of a model.fit() from a pickle file

    Args:
        filename (str): full path and name output pickle file

    Returns:
        history (dict): training history
    """

    with open(filename, 'rb') as pfile:
        history = pickle.load(pfile)

    return history

def save_datetime(value, filename):
    """saves the datetime to a pickle file

    Args:
        value (datetime):
        filename (str): full path and name output pickle file
    """
    with open(filename, 'wb') as f:
        pickle.dump(value, f)

def load_datetime(filename):
    """load the datetime from a pickle file

    Args:
        filename (str): full path and name output pickle file

    Returns:
        value (datetime): datetime
    """
    with open(filename, 'rb') as f:
        value = pickle.load(f)
    return value



import sklearn.preprocessing
import numpy as np


def make_scalers(X, method="MinMaxScaler"):
    """This function takes an input array, X, and returns
    a list of sklearn scalers for each input feature/dimension.

    Arguments:
        X {np.ndarray} -- data to make scalers for.
        The shape should be (N,M) where
        N is the number of entries or samples.
        M is the number of features/dimensionality

    Keyword Arguments:
        method {str} -- method to use to scale data.
        common choices are "MinMaxScaler" or "StandardScaler".
        But in theory any `sklearn.preprocessing` scaler method should work
         (default: {"MinMaxScaler"})

    Returns:
        list -- a least of length M of sklearn.preprocessing.[METHOD].
        The preprocessing scaler for each feature.
    """

    N, M = X.shape
    scalers = []


    # make a scaler for each feature/dimension
    for i in range(M):
        scaler_func = getattr(sklearn.preprocessing, method)()
        scalers.append(scaler_func)
        scalers[i].fit(X[:, i].reshape(-1, 1))

    return scalers


def apply_scaler(X, scalers):
    """
    applies the scalers to X
    N, M = X.shape
    N = number of entries
    M = dimensionality
    """
    N, M = X.shape

    X_scaled = np.zeros(shape=(N, M))

    for i in range(M):
        X_scaled[:, i] = scalers[i].transform(
            X[:, i].reshape(-1, 1)).reshape(1, -1)

    return X_scaled


def apply_inverse_scaler(X_scaled, scalers):
    """
    applies the inverse scalers to X_scaled
    N, M = X.shape
    N = number of entries
    M = dimensionality
    """
    N, M = X_scaled.shape

    X = np.zeros(shape=(N, M))

    for i in range(M):
        X[:, i] = scalers[i].inverse_transform(
            X_scaled[:, i].reshape(-1, 1)).reshape(1, -1)

    return X


def save_scalers(Scalers, filename):
    """save sklearn.preprocessing scalers to a .npy file

    Arguments:
        Scalers {list} -- list of sklearn.preprocessing scalers
        filename {str} -- output filename
    """

    np.save(filename, Scalers)


def load_scalers(filename, allow_pickle=True):
    """load sklearn.preprocessing scalers from a .npy file

    Arguments:
        filename {str} -- .npy file to load

    Keyword Arguments:
        allow_pickle {bool} -- allow_pickle when loading (default: {True})

    Returns:
        list -- a list of sklearn.preprocessing scalers
    """
    return np.load(filename, allow_pickle=allow_pickle)



"""
tools to explore effect of hyperparameters
"""

import tensorflow as tf
import pugna.activations


def build_model(input_dim,
                output_dim,
                units_per_layer=128,
                layers_per_block=3,
                n_blocks=3,
                activation='relu',
                leaky_alpha=0.3,
                batch_norm=False,
                summary=True):
    """
    note the way the activation functions are defined.
    This was done because when looking at model.summary()
    if you don't explicitly call the activation function
    i.e. tf.keras.layers.ReLU() or tf.keras.activations.tanh
    **IN** the for loop then it doesn't show up in the model.summary()

    Args:
        input_dim (int): number of input features
        output_dim (int): number of output features
        units_per_layer (int, optional): number of units per layers. Defaults to 128.
        layers_per_block (int, optional): number of Dense layers before a BatchNormalisation Layer is added if batch_norm is True. Defaults to 3.
        n_blocks (int, optional): number of blocks. Defaults to 3.
        activation (str, optional): specify activation function. Only the following functions are implemented: ['relu', 'leaky_relu', 'prelu', 'elu', 'softplus', 'tanh', 'selu'] Defaults to 'relu'.
        leaky_alpha (float, optional): alpha parameter for leaky_relu activation function. Defaults to 0.3.
        batch_norm (bool, optional): If True then add a BatchNormalisation Layer at the end of each 'block'. Defaults to True.
        summary (bool, optional): If True then print model.summary(). Defaults to True.

    Raises:
        ValueError: Raised if activation function supplied is not in the 'allowed_activations'

    Returns:
        keras Sequential model: a keras Sequential model ready for compilation.
    """

    allowed_activations = ['relu', 'leaky_relu',
                           'prelu', 'elu', 'softplus', 'tanh', 'selu',
                           'srelu', 'srelu2', 'srelu3']

    if activation not in allowed_activations:
        raise ValueError(
            f"activation: ({activation}) not in allowed_activations: ({allowed_activations})")

    if batch_norm:
        use_bias = False
    else:
        use_bias = True

    model = tf.keras.models.Sequential()
    model.add(tf.keras.layers.Input(shape=(input_dim,)))
    for i in range(n_blocks):
        for j in range(layers_per_block):
            #   model.add(tf.keras.layers.Dense(units_per_layer, activation=activation, use_bias=use_bias))
            # separate out the activation function so that you can load a saved model with LeakyReLU
            # https://github.com/tensorflow/tfjs/issues/1093
            model.add(tf.keras.layers.Dense(
                units_per_layer, use_bias=use_bias))
            if activation == "relu":
                model.add(tf.keras.layers.ReLU())
            elif activation == "leaky_relu":
                model.add(tf.keras.layers.LeakyReLU(alpha=leaky_alpha))
            elif activation == "prelu":
                model.add(tf.keras.layers.PReLU())
            elif activation == "elu":
                model.add(tf.keras.layers.ELU())
            elif activation == "softplus":
                model.add(tf.keras.layers.Activation(
                    tf.keras.activations.softplus))
            elif activation == "tanh":
                model.add(tf.keras.layers.Activation(
                    tf.keras.activations.tanh))
            elif activation == "selu":
                model.add(tf.keras.layers.Activation(
                    tf.keras.activations.selu))
            elif activation == "srelu":
                model.add(pugna.activations.sReLU_Layer())
            elif activation == "srelu2":
                model.add(pugna.activations.sReLU_2_Layer())
            elif activation == "srelu3":
                model.add(pugna.activations.sReLU_3_Layer())
        if batch_norm:
            model.add(tf.keras.layers.BatchNormalization())
    model.add(tf.keras.layers.Dense(output_dim, activation='linear'))
    if summary:
        model.summary()
    return model


def compile_model(model, learning_rate=1e-3, optimizer=tf.keras.optimizers.Adam, loss='mse', metrics=['mse']):
    """compile a tf.keras model

    Args:
        model (tf.keras model): A tf.keras model
        learning_rate (float, optional): initial learning rate for optimiser. Defaults to 1e-3.
        optimizer (tf.keras.optimizer, optional): Specify optimizer. Defaults to tf.keras.optimizers.Adam.
        loss (str, optional): Specify loss function. Defaults to 'mse'.
        metrics (list, optional): specify list of metrics. Defaults to ['mse'].

    Returns:
        keras model: a compiled keras model
    """
    opt = optimizer(learning_rate=learning_rate)
    model.compile(loss=loss, optimizer=opt, metrics=metrics)

    return model
